@@ -49,12 +49,7 @@
  * Author: Daniel Stone <daniel@fooishbar.org>
  */
 
-#include <stdarg.h>
-#include <stdio.h>
-#include <ctype.h>
-#include <stdlib.h>
-
-#include "xkb-priv.h"
+#include "keymap.h"
 #include "text.h"
 
 #define VMOD_HIDE_VALUE    0
@@ -132,17 +127,18 @@ err:
 static bool
 write_vmods(struct xkb_keymap *keymap, struct buf *buf)
 {
-    int num_vmods = 0;
-    int i;
+    const struct xkb_mod *mod;
+    xkb_mod_index_t num_vmods = 0;
 
-    for (i = 0; i < XkbNumVirtualMods; i++) {
-        if (!keymap->vmod_names[i])
+    darray_foreach(mod, keymap->mods) {
+        if (mod->type != MOD_VIRT)
             continue;
+
         if (num_vmods == 0)
             write_buf(buf, "\t\tvirtual_modifiers ");
         else
             write_buf(buf, ",");
-        write_buf(buf, "%s", keymap->vmod_names[i]);
+        write_buf(buf, "%s", xkb_atom_text(keymap->ctx, mod->name));
         num_vmods++;
     }
 
@@ -152,106 +148,13 @@ write_vmods(struct xkb_keymap *keymap, struct buf *buf)
     return true;
 }
 
-#define GET_TEXT_BUF_SIZE 512
-
-#define append_get_text(...) do { \
-        int _size = snprintf(ret, GET_TEXT_BUF_SIZE, __VA_ARGS__); \
-        if (_size >= GET_TEXT_BUF_SIZE) \
-            return NULL; \
-} while (0)
-
-static char *
-get_indicator_state_text(uint8_t which)
-{
-    int i;
-    static char ret[GET_TEXT_BUF_SIZE];
-    /* FIXME: Merge with ... something ... in xkbcomp? */
-    static const char *state_names[] = {
-        "base",
-        "latched",
-        "locked",
-        "effective",
-        "compat"
-    };
-
-    memset(ret, 0, GET_TEXT_BUF_SIZE);
-
-    which &= XkbIM_UseAnyMods;
-
-    if (which == 0) {
-        strcpy(ret, "0");
-        return NULL;
-    }
-
-    for (i = 0; which != 0; i++) {
-        if (!(which & (1 << i)))
-            continue;
-        which &= ~(1 << i);
-
-        if (ret[0] != '\0')
-            append_get_text("%s+%s", ret, state_names[i]);
-        else
-            append_get_text("%s", state_names[i]);
-    }
-
-    return ret;
-}
-
-static char *
-get_control_mask_text(uint32_t control_mask)
-{
-    int i;
-    static char ret[GET_TEXT_BUF_SIZE];
-    /* FIXME: Merge with ... something ... in xkbcomp. */
-    static const char *ctrl_names[] = {
-        "RepeatKeys",
-        "SlowKeys",
-        "BounceKeys",
-        "StickyKeys",
-        "MouseKeys",
-        "MouseKeysAccel",
-        "AccessXKeys",
-        "AccessXTimeout",
-        "AccessXFeedback",
-        "AudibleBell",
-        "Overlay1",
-        "Overlay2",
-        "IgnoreGroupLock"
-    };
-
-    memset(ret, 0, GET_TEXT_BUF_SIZE);
-
-    control_mask &= XkbAllBooleanCtrlsMask;
-
-    if (control_mask == 0) {
-        strcpy(ret, "none");
-        return ret;
-    }
-    else if (control_mask == XkbAllBooleanCtrlsMask) {
-        strcpy(ret, "all");
-        return ret;
-    }
-
-    for (i = 0; control_mask; i++) {
-        if (!(control_mask & (1 << i)))
-            continue;
-        control_mask &= ~(1 << i);
-
-        if (ret[0] != '\0')
-            append_get_text("%s+%s", ret, ctrl_names[i]);
-        else
-            append_get_text("%s", ctrl_names[i]);
-    }
-
-    return ret;
-}
-
 static bool
 write_keycodes(struct xkb_keymap *keymap, struct buf *buf)
 {
     struct xkb_key *key;
     struct xkb_key_alias *alias;
-    int i;
+    xkb_led_index_t i;
+    const struct xkb_indicator_map *im;
 
     if (keymap->keycodes_section_name)
         write_buf(buf, "\txkb_keycodes \"%s\" {\n",
@@ -259,31 +162,24 @@ write_keycodes(struct xkb_keymap *keymap, struct buf *buf)
     else
         write_buf(buf, "\txkb_keycodes {\n");
 
-    write_buf(buf, "\t\tminimum = %d;\n",
-              keymap->min_key_code);
-    write_buf(buf, "\t\tmaximum = %d;\n",
-              keymap->max_key_code);
-
     xkb_foreach_key(key, keymap) {
-        if (key->name[0] == '\0')
+        if (key->name == XKB_ATOM_NONE)
             continue;
 
-        write_buf(buf, "\t\t%6s = %d;\n",
-                  KeyNameText(key->name), XkbKeyGetKeycode(keymap, key));
+        write_buf(buf, "\t\t%-20s = %d;\n",
+                  KeyNameText(keymap->ctx, key->name), key->keycode);
     }
 
-    for (i = 0; i < XkbNumIndicators; i++) {
-        if (!keymap->indicator_names[i])
-            continue;
-        write_buf(buf, "\t\tindicator %d = \"%s\";\n",
-                  i + 1, keymap->indicator_names[i]);
-    }
+    darray_enumerate(i, im, keymap->indicators)
+        if (im->name != XKB_ATOM_NONE)
+            write_buf(buf, "\t\tindicator %d = \"%s\";\n",
+                      i + 1, xkb_atom_text(keymap->ctx, im->name));
 
 
     darray_foreach(alias, keymap->key_aliases)
-        write_buf(buf, "\t\talias %6s = %6s;\n",
-                  KeyNameText(alias->alias),
-                  KeyNameText(alias->real));
+        write_buf(buf, "\t\talias %-14s = %s;\n",
+                  KeyNameText(keymap->ctx, alias->alias),
+                  KeyNameText(keymap->ctx, alias->real));
 
     write_buf(buf, "\t};\n\n");
     return true;
@@ -311,7 +207,7 @@ write_types(struct xkb_keymap *keymap, struct buf *buf)
         write_buf(buf, "\t\ttype \"%s\" {\n",
                   xkb_atom_text(keymap->ctx, type->name));
         write_buf(buf, "\t\t\tmodifiers= %s;\n",
-                  VModMaskText(keymap, type->mods.mods));
+                  ModMaskText(keymap, type->mods.mods));
 
         for (j = 0; j < type->num_entries; j++) {
             const char *str;
@@ -324,7 +220,7 @@ write_types(struct xkb_keymap *keymap, struct buf *buf)
             if (entry->level == 0 && entry->preserve.mods == 0)
                 continue;
 
-            str = VModMaskText(keymap, entry->mods.mods);
+            str = ModMaskText(keymap, entry->mods.mods);
             write_buf(buf, "\t\t\tmap[%s]= Level%d;\n",
                       str, entry->level + 1);
 
@@ -332,7 +228,8 @@ write_types(struct xkb_keymap *keymap, struct buf *buf)
                 continue;
 
             write_buf(buf, "\t\t\tpreserve[%s]= ", str);
-            write_buf(buf, "%s;\n", VModMaskText(keymap, entry->preserve.mods));
+            write_buf(buf, "%s;\n",
+                      ModMaskText(keymap, entry->preserve.mods));
         }
 
         if (type->level_names) {
@@ -351,34 +248,33 @@ write_types(struct xkb_keymap *keymap, struct buf *buf)
 }
 
 static bool
-write_indicator_map(struct xkb_keymap *keymap, struct buf *buf, int num)
+write_indicator_map(struct xkb_keymap *keymap, struct buf *buf,
+                    const struct xkb_indicator_map *led)
 {
-    struct xkb_indicator_map *led = &keymap->indicators[num];
-
     write_buf(buf, "\t\tindicator \"%s\" {\n",
-              keymap->indicator_names[num]);
+              xkb_atom_text(keymap->ctx, led->name));
 
     if (led->which_groups) {
-        if (led->which_groups != XkbIM_UseEffective) {
+        if (led->which_groups != XKB_STATE_LAYOUT_EFFECTIVE) {
             write_buf(buf, "\t\t\twhichGroupState= %s;\n",
-                      get_indicator_state_text(led->which_groups));
+                      IndicatorStateText(keymap->ctx, led->which_groups));
         }
         write_buf(buf, "\t\t\tgroups= 0x%02x;\n",
                   led->groups);
     }
 
     if (led->which_mods) {
-        if (led->which_mods != XkbIM_UseEffective) {
+        if (led->which_mods != XKB_STATE_MODS_EFFECTIVE) {
             write_buf(buf, "\t\t\twhichModState= %s;\n",
-                      get_indicator_state_text(led->which_mods));
+                      IndicatorStateText(keymap->ctx, led->which_mods));
         }
         write_buf(buf, "\t\t\tmodifiers= %s;\n",
-                  VModMaskText(keymap, led->mods.mods));
+                  ModMaskText(keymap, led->mods.mods));
     }
 
     if (led->ctrls) {
         write_buf(buf, "\t\t\tcontrols= %s;\n",
-                  get_control_mask_text(led->ctrls));
+                  ControlMaskText(keymap->ctx, led->ctrls));
     }
 
     write_buf(buf, "\t\t};\n");
@@ -387,7 +283,8 @@ write_indicator_map(struct xkb_keymap *keymap, struct buf *buf, int num)
 
 static bool
 write_action(struct xkb_keymap *keymap, struct buf *buf,
-             union xkb_action *action, const char *prefix, const char *suffix)
+             const union xkb_action *action,
+             const char *prefix, const char *suffix)
 {
     const char *type;
     const char *args = NULL;
@@ -400,67 +297,68 @@ write_action(struct xkb_keymap *keymap, struct buf *buf,
     type = ActionTypeText(action->type);
 
     switch (action->type) {
-    case XkbSA_SetMods:
-    case XkbSA_LatchMods:
-    case XkbSA_LockMods:
-        if (action->mods.flags & XkbSA_UseModMapMods)
+    case ACTION_TYPE_MOD_SET:
+    case ACTION_TYPE_MOD_LATCH:
+    case ACTION_TYPE_MOD_LOCK:
+        if (action->mods.flags & ACTION_MODS_LOOKUP_MODMAP)
             args = "modMapMods";
         else
-            args = VModMaskText(keymap, action->mods.mods.mods);
+            args = ModMaskText(keymap, action->mods.mods.mods);
         write_buf(buf, "%s%s(modifiers=%s%s%s)%s", prefix, type, args,
-                  (action->type != XkbSA_LockGroup &&
-                   (action->mods.flags & XkbSA_ClearLocks)) ?
+                  (action->type != ACTION_TYPE_MOD_LOCK &&
+                   (action->mods.flags & ACTION_LOCK_CLEAR)) ?
                    ",clearLocks" : "",
-                  (action->type != XkbSA_LockGroup &&
-                   (action->mods.flags & XkbSA_LatchToLock)) ?
+                  (action->type != ACTION_TYPE_MOD_LOCK &&
+                   (action->mods.flags & ACTION_LATCH_TO_LOCK)) ?
                    ",latchToLock" : "",
                   suffix);
         break;
 
-    case XkbSA_SetGroup:
-    case XkbSA_LatchGroup:
-    case XkbSA_LockGroup:
+    case ACTION_TYPE_GROUP_SET:
+    case ACTION_TYPE_GROUP_LATCH:
+    case ACTION_TYPE_GROUP_LOCK:
         write_buf(buf, "%s%s(group=%s%d%s%s)%s", prefix, type,
-                  (!(action->group.flags & XkbSA_GroupAbsolute) &&
+                  (!(action->group.flags & ACTION_ABSOLUTE_SWITCH) &&
                    action->group.group > 0) ? "+" : "",
-                  (action->group.flags & XkbSA_GroupAbsolute) ?
+                  (action->group.flags & ACTION_ABSOLUTE_SWITCH) ?
                   action->group.group + 1 : action->group.group,
-                  (action->type != XkbSA_LockGroup &&
-                   (action->group.flags & XkbSA_ClearLocks)) ?
+                  (action->type != ACTION_TYPE_GROUP_LOCK &&
+                   (action->group.flags & ACTION_LOCK_CLEAR)) ?
                   ",clearLocks" : "",
-                  (action->type != XkbSA_LockGroup &&
-                   (action->group.flags & XkbSA_LatchToLock)) ?
+                  (action->type != ACTION_TYPE_GROUP_LOCK &&
+                   (action->group.flags & ACTION_LATCH_TO_LOCK)) ?
                   ",latchToLock" : "",
                   suffix);
         break;
 
-    case XkbSA_Terminate:
+    case ACTION_TYPE_TERMINATE:
         write_buf(buf, "%s%s()%s", prefix, type, suffix);
         break;
 
-    case XkbSA_MovePtr:
+    case ACTION_TYPE_PTR_MOVE:
         write_buf(buf, "%s%s(x=%s%d,y=%s%d%s)%s", prefix, type,
-                  (!(action->ptr.flags & XkbSA_MoveAbsoluteX) &&
+                  (!(action->ptr.flags & ACTION_ABSOLUTE_X) &&
                    action->ptr.x >= 0) ? "+" : "",
                   action->ptr.x,
-                  (!(action->ptr.flags & XkbSA_MoveAbsoluteY) &&
+                  (!(action->ptr.flags & ACTION_ABSOLUTE_Y) &&
                    action->ptr.y >= 0) ? "+" : "",
                   action->ptr.y,
-                  (action->ptr.flags & XkbSA_NoAcceleration) ? ",!accel" : "",
+                  (action->ptr.flags & ACTION_NO_ACCEL) ? ",!accel" : "",
                   suffix);
         break;
 
-    case XkbSA_LockPtrBtn:
-        switch (action->btn.flags & (XkbSA_LockNoUnlock | XkbSA_LockNoLock)) {
-        case XkbSA_LockNoUnlock:
+    case ACTION_TYPE_PTR_LOCK:
+        switch (action->btn.flags &
+                 (ACTION_LOCK_NO_LOCK | ACTION_LOCK_NO_UNLOCK)) {
+        case ACTION_LOCK_NO_UNLOCK:
             args = ",affect=lock";
             break;
 
-        case XkbSA_LockNoLock:
+        case ACTION_LOCK_NO_LOCK:
             args = ",affect=unlock";
             break;
 
-        case XkbSA_LockNoLock | XkbSA_LockNoUnlock:
+        case ACTION_LOCK_NO_LOCK | ACTION_LOCK_NO_UNLOCK:
             args = ",affect=neither";
             break;
 
@@ -468,7 +366,7 @@ write_action(struct xkb_keymap *keymap, struct buf *buf,
             args = ",affect=both";
             break;
         }
-    case XkbSA_PtrBtn:
+    case ACTION_TYPE_PTR_BUTTON:
         write_buf(buf, "%s%s(button=", prefix, type);
         if (action->btn.button > 0 && action->btn.button <= 5)
             write_buf(buf, "%d", action->btn.button);
@@ -481,43 +379,34 @@ write_action(struct xkb_keymap *keymap, struct buf *buf,
         write_buf(buf, ")%s", suffix);
         break;
 
-    case XkbSA_SetPtrDflt:
+    case ACTION_TYPE_PTR_DEFAULT:
         write_buf(buf, "%s%s(", prefix, type);
-        if (action->dflt.affect == XkbSA_AffectDfltBtn)
-            write_buf(buf, "affect=button,button=%s%d",
-                      (!(action->dflt.flags & XkbSA_DfltBtnAbsolute) &&
-                       action->dflt.value >= 0) ? "+" : "",
-                      action->dflt.value);
+        write_buf(buf, "affect=button,button=%s%d",
+                  (!(action->dflt.flags & ACTION_ABSOLUTE_SWITCH) &&
+                   action->dflt.value >= 0) ? "+" : "",
+                  action->dflt.value);
         write_buf(buf, ")%s", suffix);
         break;
 
-    case XkbSA_SwitchScreen:
+    case ACTION_TYPE_SWITCH_VT:
         write_buf(buf, "%s%s(screen=%s%d,%ssame)%s", prefix, type,
-                  (!(action->screen.flags & XkbSA_SwitchAbsolute) &&
+                  (!(action->screen.flags & ACTION_ABSOLUTE_SWITCH) &&
                    action->screen.screen >= 0) ? "+" : "",
                   action->screen.screen,
-                  (action->screen.flags & XkbSA_SwitchApplication) ? "!" : "",
+                  (action->screen.flags & ACTION_SAME_SCREEN) ? "!" : "",
                   suffix);
         break;
 
-    /* Deprecated actions below here */
-    case XkbSA_SetControls:
-    case XkbSA_LockControls:
+    case ACTION_TYPE_CTRL_SET:
+    case ACTION_TYPE_CTRL_LOCK:
         write_buf(buf, "%s%s(controls=%s)%s", prefix, type,
-                  get_control_mask_text(action->ctrls.ctrls), suffix);
+                  ControlMaskText(keymap->ctx, action->ctrls.ctrls), suffix);
         break;
 
-    case XkbSA_ISOLock:
-    case XkbSA_ActionMessage:
-    case XkbSA_RedirectKey:
-    case XkbSA_DeviceBtn:
-    case XkbSA_LockDeviceBtn:
-    case XkbSA_NoAction:
-        /* XXX TODO */
+    case ACTION_TYPE_NONE:
         write_buf(buf, "%sNoAction()%s", prefix, suffix);
         break;
 
-    case XkbSA_XFree86Private:
     default:
         write_buf(buf,
                   "%s%s(type=0x%02x,data[0]=0x%02x,data[1]=0x%02x,data[2]=0x%02x,data[3]=0x%02x,data[4]=0x%02x,data[5]=0x%02x,data[6]=0x%02x)%s",
@@ -535,8 +424,8 @@ write_action(struct xkb_keymap *keymap, struct buf *buf,
 static bool
 write_compat(struct xkb_keymap *keymap, struct buf *buf)
 {
-    int i;
     struct xkb_sym_interpret *interp;
+    const struct xkb_indicator_map *led;
 
     if (keymap->compat_section_name)
         write_buf(buf, "\txkb_compatibility \"%s\" {\n\n",
@@ -548,9 +437,8 @@ write_compat(struct xkb_keymap *keymap, struct buf *buf)
 
     write_buf(buf, "\t\tinterpret.useModMapMods= AnyLevel;\n");
     write_buf(buf, "\t\tinterpret.repeat= False;\n");
-    write_buf(buf, "\t\tinterpret.locking= False;\n");
 
-    darray_foreach(interp, keymap->sym_interpret) {
+    darray_foreach(interp, keymap->sym_interprets) {
         char keysym_name[64];
 
         if (interp->sym == XKB_KEY_NoSymbol)
@@ -561,43 +449,25 @@ write_compat(struct xkb_keymap *keymap, struct buf *buf)
         write_buf(buf, "\t\tinterpret %s+%s(%s) {\n",
                   keysym_name,
                   SIMatchText(interp->match),
-                  VModMaskText(keymap, interp->mods));
+                  ModMaskText(keymap, interp->mods));
 
-        if (interp->virtual_mod != XkbNoModifier) {
+        if (interp->virtual_mod != XKB_MOD_INVALID)
             write_buf(buf, "\t\t\tvirtualModifier= %s;\n",
-                      keymap->vmod_names[interp->virtual_mod]);
-        }
+                      ModIndexText(keymap, interp->virtual_mod));
 
-        if (interp->match & XkbSI_LevelOneOnly)
-            write_buf(buf,
-                      "\t\t\tuseModMapMods=level1;\n");
-        if (interp->flags & XkbSI_LockingKey)
-            write_buf(buf, "\t\t\tlocking= True;\n");
-        if (interp->flags & XkbSI_AutoRepeat)
+        if (interp->level_one_only)
+            write_buf(buf, "\t\t\tuseModMapMods=level1;\n");
+        if (interp->repeat)
             write_buf(buf, "\t\t\trepeat= True;\n");
 
-        write_action(keymap, buf, &interp->act, "\t\t\taction= ", ";\n");
+        write_action(keymap, buf, &interp->action, "\t\t\taction= ", ";\n");
         write_buf(buf, "\t\t};\n");
     }
 
-    for (i = 0; i < XkbNumKbdGroups; i++) {
-        struct xkb_mods *gc;
-
-        gc = &keymap->groups[i];
-        if (gc->mods == 0)
-            continue;
-        write_buf(buf, "\t\tgroup %d = %s;\n", i + 1,
-                  VModMaskText(keymap, gc->mods));
-    }
-
-    for (i = 0; i < XkbNumIndicators; i++) {
-        struct xkb_indicator_map *map = &keymap->indicators[i];
-        if (map->flags == 0 && map->which_groups == 0 &&
-            map->groups == 0 && map->which_mods == 0 &&
-            map->mods.mods == 0 && map->ctrls == 0)
-            continue;
-        write_indicator_map(keymap, buf, i);
-    }
+    darray_foreach(led, keymap->indicators)
+        if (led->which_groups || led->groups || led->which_mods ||
+            led->mods.mods || led->ctrls)
+            write_indicator_map(keymap, buf, led);
 
     write_buf(buf, "\t};\n\n");
 
@@ -606,7 +476,7 @@ write_compat(struct xkb_keymap *keymap, struct buf *buf)
 
 static bool
 write_keysyms(struct xkb_keymap *keymap, struct buf *buf,
-              struct xkb_key *key, xkb_group_index_t group)
+              struct xkb_key *key, xkb_layout_index_t group)
 {
     const xkb_keysym_t *syms;
     int num_syms;
@@ -614,11 +484,11 @@ write_keysyms(struct xkb_keymap *keymap, struct buf *buf,
 #define OUT_BUF_LEN 128
     char out_buf[OUT_BUF_LEN];
 
-    for (level = 0; level < XkbKeyGroupWidth(keymap, key, group); level++) {
+    for (level = 0; level < XkbKeyGroupWidth(key, group); level++) {
         if (level != 0)
             write_buf(buf, ", ");
-        num_syms = xkb_key_get_syms_by_level(keymap, key, group, level,
-                                             &syms);
+        num_syms = xkb_keymap_key_get_syms_by_level(keymap, key->keycode,
+                                                    group, level, &syms);
         if (num_syms == 0) {
             write_buf(buf, "%15s", "NoSymbol");
         }
@@ -633,7 +503,7 @@ write_keysyms(struct xkb_keymap *keymap, struct buf *buf,
                 if (s != 0)
                     write_buf(buf, ", ");
                 xkb_keysym_get_name(syms[s], out_buf, OUT_BUF_LEN);
-                write_buf(buf, "%15s", out_buf);
+                write_buf(buf, "%s", out_buf);
             }
             write_buf(buf, " }");
         }
@@ -647,7 +517,7 @@ static bool
 write_symbols(struct xkb_keymap *keymap, struct buf *buf)
 {
     struct xkb_key *key;
-    xkb_group_index_t group, tmp;
+    xkb_layout_index_t group;
     bool showActions;
 
     if (keymap->symbols_section_name)
@@ -656,55 +526,55 @@ write_symbols(struct xkb_keymap *keymap, struct buf *buf)
     else
         write_buf(buf, "\txkb_symbols {\n\n");
 
-    for (tmp = group = 0; group < XkbNumKbdGroups; group++) {
-        if (!keymap->group_names[group])
-            continue;
-        write_buf(buf,
-                  "\t\tname[group%d]=\"%s\";\n", group + 1,
-                  keymap->group_names[group]);
-        tmp++;
-    }
-    if (tmp > 0)
+    for (group = 0; group < keymap->num_group_names; group++)
+        if (keymap->group_names[group])
+            write_buf(buf,
+                      "\t\tname[group%d]=\"%s\";\n", group + 1,
+                      xkb_atom_text(keymap->ctx, keymap->group_names[group]));
+    if (group > 0)
         write_buf(buf, "\n");
 
     xkb_foreach_key(key, keymap) {
         bool simple = true;
+        bool explicit_types = false;
+        bool multi_type = false;
 
         if (key->num_groups == 0)
             continue;
 
-        write_buf(buf, "\t\tkey %6s {", KeyNameText(key->name));
+        write_buf(buf, "\t\tkey %-20s {", KeyNameText(keymap->ctx, key->name));
 
-        if (key->explicit & XkbExplicitKeyTypesMask) {
-            bool multi_type = false;
-            struct xkb_key_type *type = XkbKeyType(keymap, key, 0);
+        for (group = 0; group < key->num_groups; group++) {
+            if (key->groups[group].explicit_type)
+                explicit_types = true;
 
+            if (group != 0 && key->groups[group].type != key->groups[0].type)
+                multi_type = true;
+        }
+
+        if (explicit_types) {
+            const struct xkb_key_type *type;
             simple = false;
 
-            for (group = 1; group < key->num_groups; group++) {
-                if (XkbKeyType(keymap, key, group) != type) {
-                    multi_type = true;
-                    break;
-                }
-            }
-
             if (multi_type) {
                 for (group = 0; group < key->num_groups; group++) {
-                    if (!(key->explicit & (1 << group)))
+                    if (!key->groups[group].explicit_type)
                         continue;
-                    type = XkbKeyType(keymap, key, group);
+
+                    type = key->groups[group].type;
                     write_buf(buf, "\n\t\t\ttype[group%u]= \"%s\",",
                               group + 1,
                               xkb_atom_text(keymap->ctx, type->name));
                 }
             }
             else {
+                type = key->groups[0].type;
                 write_buf(buf, "\n\t\t\ttype= \"%s\",",
                           xkb_atom_text(keymap->ctx, type->name));
             }
         }
 
-        if (key->explicit & XkbExplicitAutoRepeatMask) {
+        if (key->explicit & EXPLICIT_REPEAT) {
             if (key->repeats)
                 write_buf(buf, "\n\t\t\trepeat= Yes,");
             else
@@ -712,27 +582,25 @@ write_symbols(struct xkb_keymap *keymap, struct buf *buf)
             simple = false;
         }
 
-        if (key->vmodmap && (key->explicit & XkbExplicitVModMapMask)) {
-            /* XXX: vmodmap cmask? */
+        if (key->vmodmap && (key->explicit & EXPLICIT_VMODMAP))
             write_buf(buf, "\n\t\t\tvirtualMods= %s,",
-                      VModMaskText(keymap, key->vmodmap << XkbNumModifiers));
-        }
+                      ModMaskText(keymap, key->vmodmap));
 
         switch (key->out_of_range_group_action) {
-        case XkbClampIntoRange:
+        case RANGE_SATURATE:
             write_buf(buf, "\n\t\t\tgroupsClamp,");
             break;
 
-        case XkbRedirectIntoRange:
+        case RANGE_REDIRECT:
             write_buf(buf, "\n\t\t\tgroupsRedirect= Group%u,",
                       key->out_of_range_group_number + 1);
             break;
+
+        default:
+            break;
         }
 
-        if (key->explicit & XkbExplicitInterpretMask)
-            showActions = (key->actions != NULL);
-        else
-            showActions = false;
+        showActions = !!(key->explicit & EXPLICIT_INTERP);
 
         if (key->num_groups > 1 || showActions)
             simple = false;
@@ -757,12 +625,11 @@ write_symbols(struct xkb_keymap *keymap, struct buf *buf)
                     write_buf(buf, ",\n\t\t\tactions[Group%u]= [ ",
                               group + 1);
                     for (level = 0;
-                         level < XkbKeyGroupWidth(keymap, key, group);
-                         level++) {
+                         level < XkbKeyGroupWidth(key, group); level++) {
                         if (level != 0)
                             write_buf(buf, ", ");
                         write_action(keymap, buf,
-                                     XkbKeyActionEntry(key, group, level),
+                                     &key->groups[group].levels[level].action,
                                      NULL, NULL);
                     }
                     write_buf(buf, " ]");
@@ -773,17 +640,19 @@ write_symbols(struct xkb_keymap *keymap, struct buf *buf)
     }
 
     xkb_foreach_key(key, keymap) {
-        int mod;
+        xkb_mod_index_t i;
+        const struct xkb_mod *mod;
 
         if (key->modmap == 0)
             continue;
 
-        for (mod = 0; mod < XkbNumModifiers; mod++) {
-            if (!(key->modmap & (1 << mod)))
+        darray_enumerate(i, mod, keymap->mods) {
+            if (!(key->modmap & (1 << i)))
                 continue;
 
             write_buf(buf, "\t\tmodifier_map %s { %s };\n",
-                      ModIndexToName(mod), KeyNameText(key->name));
+                      xkb_atom_text(keymap->ctx, mod->name),
+                      KeyNameText(keymap->ctx, key->name));
         }
     }
 
@@ -792,11 +661,22 @@ write_symbols(struct xkb_keymap *keymap, struct buf *buf)
 }
 
 XKB_EXPORT char *
-xkb_map_get_as_string(struct xkb_keymap *keymap)
+xkb_keymap_get_as_string(struct xkb_keymap *keymap,
+                         enum xkb_keymap_format format)
 {
     bool ok;
     struct buf buf = { NULL, 0, 0 };
 
+    if (format == XKB_KEYMAP_USE_ORIGINAL_FORMAT)
+        format = keymap->format;
+
+    if (format != XKB_KEYMAP_FORMAT_TEXT_V1) {
+        log_err(keymap->ctx,
+                "Trying to get a keymap as a string in an unsupported format (%d)\n",
+                format);
+        return NULL;
+    }
+
     ok = (check_write_buf(&buf, "xkb_keymap {\n") &&
           write_keycodes(keymap, &buf) &&
           write_types(keymap, &buf) &&
